@@ -2,7 +2,7 @@
 
 ;; Copyright (C) 2015-2017  Jorgen Schaefer <contact@jorgenschaefer.de>
 
-;; Version: 1.26
+;; Version: 1.31
 ;; Author: Jorgen Schaefer <contact@jorgenschaefer.de>
 ;; Maintainer: Ola Nilsson <ola.nilsson@gmail.com>
 ;; Package-Requires: ((emacs "24.3"))
@@ -50,10 +50,19 @@
 
 ;; A base error for all errors raised by buttercup.
 (define-error 'buttercup-error-base "error")
+;; Buttercup internals error, raised on internal implementation
+;; inconsistencies.
+(define-error 'buttercup-internals-error "Internal buttercup error"
+              'buttercup-error-base)
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; wrapper function manipulation
 
+;; Error for buttercup--enclosed-expr
+(define-error 'buttercup-enclosed-expression-error
+              "Bad test expression"
+              'buttercup-internals-error)
+
 (defun buttercup--enclosed-expr (fun)
   "Given a zero-arg function FUN, return its unevaluated expression.
 
@@ -76,18 +85,14 @@ ensures access to the un-expanded form."
     (`(closure ,(pred listp) nil
         (quote ,expr) (buttercup--mark-stackframe) ,_expanded)
      expr)
-    ;; This a when FUN has not been evaluated. Probably never happens
-    ;; except when testing buttercup. Should probably do something
-    ;; about that.
+    ;; This a when FUN has not been evaluated.
+    ;; Why does that happen?
     ;; A lambda with an empty arglist and a body containing
     ;; * the quoted original expression
     ;; * the stackframe marker
-    ;; * the original expression
-    ;; In this case expr and expr2 should be equal (but not eq?) as
-    ;; expr2 has not been macroexpanded.
-    ((and `(lambda nil
-             (quote ,expr) (buttercup--mark-stackframe) ,expr2)
-          (guard (equal expr expr2)))
+    ;; * the expanded expression
+    (`(lambda nil
+        (quote ,expr) (buttercup--mark-stackframe) ,_expanded)
      expr)
     ;;; This is when FUN has been byte compiled, as when the entire
     ;;; test file has been byte compiled. Check that it has an empty
@@ -97,7 +102,7 @@ ensures access to the un-expanded form."
     ((and (pred byte-code-function-p) (guard (member (aref fun 0) '(nil 0))))
      (aref fun 1))
     ;; Error
-    (_ (error "Not a zero-arg one-expression closure: %S" fun))))
+    (_ (signal 'buttercup-enclosed-expression-error (format "Not a zero-arg one-expression closure: %S" fun)))))
 
 (defun buttercup--expr-and-value (fun)
   "Given a function, return its quoted expression and value.
@@ -129,6 +134,34 @@ a call to `save-match-data', as `format-spec' modifies that."
   (save-match-data
     (format-spec format specification)))
 
+(defun buttercup--simple-format (specification &rest format)
+  "Return a string based on SPECIFICATION and FORMAT.
+A simpler version of `format-spec', which see.
+If more than one FORMAT string is given they will be combined
+before formatting replacements occur.
+
+Does not support flags, width or precision.
+The substitution for a specification character can be a function,
+which is only supported in `format-spec' from Emacs 29.
+Does not have the IGNORE-MISSING and SPLIT parameters."
+  (save-match-data
+    (with-temp-buffer
+      (apply #'insert format)
+      (goto-char 1)
+      (while (search-forward "%" nil t)
+        (cond
+         ((= (following-char) ?%)
+          (delete-char 1))
+         ((looking-at-p (rx alpha))
+          (let* ((char (following-char))
+                 (begin (point))
+                 (replacement (cdr (assq char specification)))
+                 (text (if (functionp replacement) (funcall replacement) replacement)))
+            (insert-and-inherit text)
+            (delete-char 1)
+            (delete-region (1- begin) begin)))))
+      (buffer-string))))
+
 ;;;;;;;;;;
 ;;; expect
 
@@ -221,7 +254,8 @@ The BODY will receive ARGS as functions that can be called (using
 simple boolean, or a cons cell of the form (RESULT . MESSAGE). If
 RESULT is nil, MESSAGE should describe why the matcher failed. If
 RESULT is non-nil, MESSAGE should describe why a negated matcher
-failed."
+failed.
+BODY may start with a docstring."
   (declare (indent defun))
   `(put ,matcher 'buttercup-matcher
         (lambda ,args
@@ -345,7 +379,7 @@ See also `buttercup-define-matcher'."
                  (concat expect-match-phrase " "
                          expect-mismatch-phrase))))
       (if (symbolp function)
-          (setq function-name function)
+          (setq function-name (symbol-name function))
         (error "The `:function-name' keyword is required if FUNCTION is not a symbol")))
     `(buttercup-define-matcher ,matcher (arg)
        (let* ((expr (buttercup--enclosed-expr arg))
@@ -421,7 +455,7 @@ See also `buttercup-define-matcher'."
                  (concat expect-match-phrase " "
                          expect-mismatch-phrase))))
       (if (symbolp function)
-          (setq function-name function)
+          (setq function-name (symbol-name function))
         (error "The `:function-name' keyword is required if FUNCTION is not a symbol")))
     `(buttercup-define-matcher ,matcher (a b)
        (cl-destructuring-bind
@@ -493,7 +527,7 @@ See also `buttercup-define-matcher'."
       ((text-expr . text) (regexp-expr . regexp))
       (mapcar #'buttercup--expr-and-value (list text regexp))
     (save-match-data
-      (let* (;; For string literals, juse use them normally, but for
+      (let* (;; For string literals, just use them normally, but for
              ;; expressions, show both the expr and its string value
              (text-is-literal (equal text-expr text))
              (regexp-is-literal (equal regexp-expr regexp))
@@ -578,13 +612,19 @@ See also `buttercup-define-matcher'."
                 a-expr b tolerance a (abs (- a b)))))))
 
 (buttercup-define-matcher :to-throw (expr &optional signal signal-args)
-  (let ((expected-signal-symbol (or (and signal (funcall signal)) 'error))
+  "Check that EXPR raises SIGNAL with SIGNAL-ARGS.
+EXPR, SIGNAL, and SIGNAL-ARGS should all be buttercup-wrapped
+objects. EXPR is the test code that will be evaluated. The signal
+raised by EXPR must be SIGNAL or an error signal derived from
+SIGNAL. If SIGNAL is nil match any error signal. If SIGNAL-ARGS
+are given they must be `equal' to the arguments of the caught
+signal. Do not consider the signal arguments if SIGNAL-ARGS is
+nil."
+  (let ((expected-signal-symbol (and signal (funcall signal)))
         (expected-signal-args (and signal-args (funcall signal-args)))
         (unevaluated-expr (buttercup--enclosed-expr expr))
         expr-value
-        thrown-signal
-        thrown-signal-symbol
-        thrown-signal-args)
+        thrown-signal)
     (when (and (functionp unevaluated-expr)
                (member (car unevaluated-expr) '(lambda closure)))
       (display-warning
@@ -593,19 +633,33 @@ See also `buttercup-define-matcher'."
         (format "Probable incorrect use of `:to-throw' matcher: pass an expression instead of a function: `%S'"
                 unevaluated-expr)
         'yellow)))
-    ;; If no signal specificaiton, use `error' as the signal symbol
-    (when (and (null expected-signal-symbol)
-               (null expected-signal-args))
-      (setq expected-signal-symbol 'error))
-    ;; Set the above 4 variables
+    ;; Set the above variables
     (condition-case err
         (setq expr-value
               (funcall expr))
       (error
-       (setq thrown-signal err
-             thrown-signal-symbol (car err)
-             thrown-signal-args (cdr err))
+       (setq thrown-signal err)
        nil))
+    (buttercup--handle-to-throw thrown-signal
+                                (cons expected-signal-symbol expected-signal-args)
+                                unevaluated-expr expr-value)))
+
+(defun buttercup--handle-to-throw (thrown-signal expected-signal unevaluated-expr expr-value)
+  "Handle the results of the :to-throw matcher.
+This is a separate function for testability purposes.
+THROWN-SIGNAL is the signal - a `cons' of symbol and arguments -
+caught by `condition-case', or nil if no signal was raised.
+EXPECTED-SIGNAL is a `cons' of the expected signal symbol and
+arguments. The `cdr' can be nil if the `expect' statement did not
+specify any expected arguments.
+UNEVALUATED-EXPR is the Lisp sexp used before the :to-throw
+matcher keyword in the `expect' statement.
+EXPR-VALUE is the return value from the evaluation of
+UNEVALUATED-EXPR if it did not raise any signal."
+  (let ((thrown-signal-symbol (car thrown-signal))
+        (thrown-signal-args (cdr thrown-signal))
+        (expected-signal-symbol (car expected-signal))
+        (expected-signal-args (cdr expected-signal)))
     (let*
         ((matched
           (and thrown-signal
@@ -613,36 +667,38 @@ See also `buttercup-define-matcher'."
                    (memq expected-signal-symbol (get thrown-signal-symbol 'error-conditions)))
                (or (null expected-signal-args)
                    (equal thrown-signal-args expected-signal-args))))
+         ;; Some of these replacement are always used, there is no
+         ;; reason not to format them immediately. But e and t are not
+         ;; always used and should be delayed. Use
+         ;; buttercup--simple-format for formatting as format-spec
+         ;; does not support functions until Emacs 29
          (spec (format-spec-make
                 ?E (format "%S" unevaluated-expr)
-                ?e (format "%S" expr-value)
-                ?t (format "%S" thrown-signal)
+                ?e (lambda () (format "%S" expr-value))
+                ?t (lambda () (format "%S" thrown-signal))
+                ?S (lambda () (format "%S" thrown-signal-symbol))
+                ?A (lambda () (format "%S" thrown-signal-args))
                 ?s (if expected-signal-symbol
                        (format "a child signal of `%S'" expected-signal-symbol)
                      "a signal")
                 ?a (if expected-signal-args
                        (format " with args `%S'" expected-signal-args)
                      "")))
-         (result-text
-          (if thrown-signal
-              (buttercup-format-spec "it threw %t" spec)
-            (buttercup-format-spec "it evaluated successfully, returning value `%e'" spec)))
-
-         (expect-match-text
-          (concat (buttercup-format-spec "Expected `%E' to throw %s%a" spec)
-                  ", but instead "
-                  result-text))
-         (expect-mismatch-text
-          (concat (buttercup-format-spec "Expected `%E' not to throw %s%a" spec)
-                  ", but "
-                  result-text)))
+         (result-fmt (if thrown-signal
+                         "it threw `%S' with args %A"
+                       "it evaluated successfully, returning value `%e'")))
       (buttercup--test-expectation matched
-        :expect-match-phrase expect-match-text
-        :expect-mismatch-phrase expect-mismatch-text))))
+        :expect-match-phrase
+        (buttercup--simple-format
+         spec
+         "Expected `%E' to throw %s%a, but instead " result-fmt)
+        :expect-mismatch-phrase
+        (buttercup--simple-format
+         spec
+         "Expected `%E' not to throw %s%a, but " result-fmt)))))
 
 (buttercup-define-matcher :to-have-been-called (spy)
-  (setq spy (funcall spy))
-  (cl-assert (symbolp spy))
+  (cl-assert (symbolp (funcall spy)))
   (if (spy-calls-all (funcall spy))
       t
     nil))
@@ -863,7 +919,10 @@ mainly calls to `describe', `it' and `before-each'."
            `((let* ,(elt body 1)
                ,@(cddr body))))
           (t body))))
-    `(buttercup-describe ,description (lambda () ,@new-body))))
+    (if (or (memq :var new-body)
+              (memq :var* new-body))
+      `(error "buttercup: :var(*) found in invalid position of describe form \"%s\"" ,description)
+    `(buttercup-describe ,description (lambda () ,@new-body)))))
 
 (defun buttercup-describe (description body-function)
   "Function to handle a `describe' form.
@@ -1116,23 +1175,18 @@ none of them is equivalent to `:return-value nil'.
 
 (defun spy-on (symbol &optional keyword arg)
   "Create a spy (mock) for the function SYMBOL.
+The spy will track call information that can be queried with the
+:to-have-been-called matchers.
+The default spy will always return nil, but that can be changed
+by setting KEYWORD to one of:
 
-KEYWORD can have one of the following values:
+  :and-call-through -- call the original function.
 
-  :and-call-through -- Track calls, but call the original
-      function.
+  :and-return-value -- return ARG.
 
-  :and-return-value -- Track calls, but return ARG instead of
-      calling the original function.
+  :and-call-fake -- call ARG, a function that matches SYMBOL.
 
-  :and-call-fake -- Track calls, but call ARG instead of the
-      original function.
-
-  :and-throw-error -- Signal ARG as an error instead of calling
-      the original function.
-
-  nil -- Track calls, but simply return nil instead of calling
-      the original function.
+  :and-throw-error -- Signal ARG without arguments.
 
 If the original function was a command, the generated spy will
 also be a command with the same interactive form, unless
@@ -1182,7 +1236,7 @@ responsibility to ensure ARG is a command."
             (_
              (error "Invalid `spy-on' keyword: `%S'" keyword)))))
     (unless (buttercup--spy-on-and-call-replacement symbol replacement)
-      (error "Spies can only be created in `before-each'"))))
+      (error "Spies can only be created in `before-each' or `it'"))))
 
 
 (defun buttercup--spy-on-and-call-replacement (spy fun)
@@ -1235,13 +1289,24 @@ responsibility to ensure ARG is a command."
 Should always be set to a value that is not `listp', except while
 in a `buttercup-with-cleanup' environment.")
 
+(defvar native-comp-enable-subr-trampolines)
+(defvar comp-enable-subr-trampolines)
+
 (defmacro buttercup-with-cleanup (&rest body)
   "Execute BODY, cleaning spys and the rest afterwards."
-  `(let ((buttercup--cleanup-functions nil))
-     (unwind-protect (progn ,@body)
-       (dolist (fun buttercup--cleanup-functions)
-         (ignore-errors
-           (funcall fun))))))
+  `(,@(if (fboundp 'with-suppressed-warnings)
+          '(with-suppressed-warnings ((obsolete comp-enable-subr-trampolines)))
+        '(progn))
+     (let ((buttercup--cleanup-functions nil)
+           ;; Redefining certain primitive's trampolines will cause problems,
+           ;; see https://github.com/jorgenschaefer/emacs-buttercup/issues/230 and
+           ;; https://debbugs.gnu.org/cgi/bugreport.cgi?bug=61880
+           (comp-enable-subr-trampolines nil)
+           (native-comp-enable-subr-trampolines nil))
+       (unwind-protect (progn ,@body)
+         (dolist (fun buttercup--cleanup-functions)
+           (ignore-errors
+             (funcall fun)))))))
 
 (defun buttercup--add-cleanup (function)
   "Register FUNCTION for cleanup in `buttercup-with-cleanup'."
@@ -1268,11 +1333,6 @@ in a `buttercup-with-cleanup' environment.")
            nil
            buttercup--spy-contexts))
 
-(buttercup-define-matcher :to-have-been-called (spy)
-  (if (spy-calls-all (funcall spy))
-      t
-    nil))
-
 (defun spy-calls-any (spy)
   "Return t iff SPY has been called at all, nil otherwise."
   (if (spy-calls-all spy)
@@ -1921,7 +1981,7 @@ the capturing behavior."
   "Format STRING with COLOR.
 Return STRING unmodified if COLOR is nil."
   (if color
-      (let ((color-code (cdr (assoc color buttercup-colors))))
+      (let ((color-code (cdr (assq color buttercup-colors))))
         (format "\e[%sm%s\e[0m" color-code string))
     string))
 
@@ -2086,6 +2146,7 @@ With buttercup minor mode active the following is activated:
 - `describe' and `it' forms are available from `imenu' for
   quicker access."
   :lighter " ❀"
+  :keymap (make-sparse-keymap)
   (let ((font-lock-form '(("(\\(describe\\|buttercup-define-matcher\\|it\\) "
                            1 'font-lock-keyword-face)))
         (imenu-forms '(("Test Suites" "\\((describe\\_> +\\)\"\\(\\_<.+\\_>\\)\"" 2)
@@ -2125,9 +2186,11 @@ ARG is ‘toggle’; disable otherwise."
            (add-hook 'after-load-functions #'buttercup-check-for-stale-elc)))
         (t (remove-hook 'after-load-functions #'buttercup-check-for-stale-elc))))
 
+(provide 'buttercup)
+;;; buttercup.el ends here
+
 ;; Local Variables:
 ;; indent-tabs-mode: nil
 ;; sentence-end-double-space: nil
+;; tab-width: 8
 ;; End:
-(provide 'buttercup)
-;;; buttercup.el ends here
